package org.beigesoft.handler;

/*
 * Copyright (c) 2017 Beigesoft™
 *
 * Licensed under the GNU General Public License (GPL), Version 2.0
 * (the "License");
 * you may not use this file except in compliance with the License.
 *
 * You may obtain a copy of the License at
 *
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.en.html
 */

import java.util.Map;

import org.beigesoft.exception.ExceptionWithCode;
import org.beigesoft.model.IRequestData;
import org.beigesoft.model.IHasId;
import org.beigesoft.log.ILogger;
import org.beigesoft.delegate.IDelegate;
import org.beigesoft.service.IFillerObjectsFrom;
import org.beigesoft.service.IProcessor;
import org.beigesoft.service.IEntityProcessor;
import org.beigesoft.factory.IFactorySimple;
import org.beigesoft.factory.IFactoryAppBeansByName;
import org.beigesoft.factory.IFactoryAppBeansByClass;
import org.beigesoft.holder.IHolderForClassByName;
import org.beigesoft.service.ISrvDatabase;

/**
 * <p>Handler for entity requests like "create, update, delete, list".
 * According "Beigesoft business logic programming".
 * All requests requires transaction.</p>
 *
 * @param <RS> platform dependent RDBMS recordset
 * @author Yury Demidenko
 */
public class HandlerEntityRequest<RS> implements IHandlerRequest {

  /**
   * <p>Logger.</p>
   **/
  private ILogger logger;

  /**
   * <p>Security logger.</p>
   **/
  private ILogger secureLogger;

  /**
   * <p>Database service.</p>
   */
  private ISrvDatabase<RS> srvDatabase;

  /**
   * <p>Service that fill entity from request.</p>
   **/
  private IFillerObjectsFrom<IRequestData> fillEntityFromReq;

  /**
   * <p>Entities factories factory.</p>
   **/
  private IFactoryAppBeansByClass<IFactorySimple<?>> entitiesFactoriesFatory;

  /**
   * <p>Entities processors factory.</p>
   **/
  private IFactoryAppBeansByName<IEntityProcessor> entitiesProcessorsFactory;

  /**
   * <p>Entities processors names holder.</p>
   **/
  private IHolderForClassByName<String> entitiesProcessorsNamesHolder;

  /**
   * <p>Processors factory.</p>
   **/
  private IFactoryAppBeansByName<IProcessor> processorsFactory;

  /**
   * <p>Processors names holder.</p>
   **/
  private IHolderForClassByName<String> processorsNamesHolder;

  /**
   * <p>Entities map "EntitySimpleName"-"Class".</p>
   **/
  private Map<String, Class<?>> entitiesMap;

  /**
   * <p>Transaction isolation for changing DB phase.</p>
   **/
  private Integer writeTi = ISrvDatabase.TRANSACTION_READ_UNCOMMITTED;

  /**
   * <p>Transaction isolation for reading DB phase.</p>
   **/
  private Integer readTi = ISrvDatabase.TRANSACTION_READ_UNCOMMITTED;

  /**
   * <p>Transaction isolation for writing and reading DB phase.</p>
   **/
  private Integer writeReTi = ISrvDatabase.TRANSACTION_READ_UNCOMMITTED;

  /**
   * <p>Writing and reading phases in separated transactions.</p>
   **/
  private Boolean wrReSpTr = Boolean.FALSE;

  /**
   * <p>To assemble service with additional business logic
   * (it's like on start interceptor).</p>
   **/
  private IDelegate<IRequestData> delegateSt;

  /**
   * <p>Handle request.
   * actions that change database use read_commited TI
   * this usually required only for complex transactions like loading
   * or withdrawal warehouse (receipt or issue).
   * WHandlerAndJsp requires handle NULL request, so if parameter
   * "nmEnt" is null then do nothing.
   * </p>
   * @param pRqVs Request scoped variables
   * @param pRqDt Request Data
   * @throws Exception - an exception
   */
  @Override
  public final void handle(final Map<String, Object> pRqVs,
    final IRequestData pRqDt) throws Exception {
    String nmEnt = pRqDt.getParameter("nmEnt");
    if (nmEnt == null) {
      //WHandlerAndJsp requires handle NULL request:
      return;
    }
    Class<?> entityClass = this.entitiesMap.get(nmEnt);
    if (entityClass == null) {
      this.secureLogger.error(null, HandlerEntityRequest.class,
    "Trying to work with forbidden entity/host/addr/port/user: " + nmEnt + "/"
  + pRqDt.getRemoteHost() + "/" + pRqDt.getRemoteAddr() + "/"
+ pRqDt.getRemotePort() + "/" + pRqDt.getUserName());
      throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
        "Forbidden!");
    }
    if (this.delegateSt != null) {
      this.delegateSt.makeWith(pRqVs, pRqDt);
    }
    int detLev = this.logger.getDetailLevel();
    boolean isShowDebMsg = this.logger.getIsShowDebugMessages()
      && this.logger.getIsShowDebugMessagesFor(this.getClass());
    String[] actionsArr = pRqDt.getParameter("nmsAct").split(",");
    for (String actionNm : actionsArr) {
      if (isShowDebMsg && detLev > 100) {
        this.logger.debug(pRqVs, HandlerEntityRequest.class,
          "Action: " + actionNm);
      }
    }
    if (this.wrReSpTr && ("entityFolSave".equals(actionsArr[0])
  || "entitySave".equals(actionsArr[0]) || "entityDelete".equals(actionsArr[0])
|| "entityFolDelete".equals(actionsArr[0]))) {
      IHasId<?> entity = null;
      try {
        this.srvDatabase.setIsAutocommit(false);
        this.srvDatabase.setTransactionIsolation(this.writeTi);
        this.srvDatabase.beginTransaction();
        @SuppressWarnings("unchecked")
        IFactorySimple<IHasId<?>> entFac = (IFactorySimple<IHasId<?>>)
          this.entitiesFactoriesFatory.lazyGet(pRqVs, entityClass);
        entity = entFac.create(pRqVs);
        this.fillEntityFromReq.fill(pRqVs, entity, pRqDt);
        String entProcNm = this.entitiesProcessorsNamesHolder
          .getFor(entityClass, actionsArr[0]);
        if (entProcNm == null) {
          this.secureLogger.error(null, HandlerEntityRequest.class,
            "Trying to work with forbidden entity/action/user: " + nmEnt + "/"
              + actionsArr[0] + "/" + pRqDt.getUserName());
          throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
            "Forbidden!");
        }
        @SuppressWarnings("unchecked")
        IEntityProcessor<IHasId<?>, ?> ep = (IEntityProcessor<IHasId<?>, ?>)
          this.entitiesProcessorsFactory.lazyGet(pRqVs, entProcNm);
        if (isShowDebMsg && detLev > 100) {
          this.logger.debug(pRqVs, HandlerEntityRequest.class,
            "CHANGING transaction use entProcNm/IEntityProcessor: " + entProcNm
              + "/" + ep.getClass());
        }
        entity = ep.process(pRqVs, entity, pRqDt);
        this.srvDatabase.commitTransaction();
      } catch (Exception ex) {
        this.srvDatabase.rollBackTransaction();
        throw ex;
      } finally {
        this.srvDatabase.releaseResources();
      }
      if (actionsArr.length > 1) { //reading phase:
        try {
          this.srvDatabase.setIsAutocommit(false);
          this.srvDatabase.setTransactionIsolation(this.readTi);
          this.srvDatabase.beginTransaction();
          for (int i = 1; i < actionsArr.length; i++) {
            String actionNm = actionsArr[i];
            if (actionNm.startsWith("entity")) {
              if (entity == null) { // it's may be change entity to owner:
                entity = (IHasId<?>) pRqVs.get("nextEntity");
                if (entity == null) {
                  throw new ExceptionWithCode(ExceptionWithCode.WRONG_PARAMETER,
                    "wrong_request_entity_not_filled");
                }
               entityClass = entity.getClass();
              }
              String entProcNm = this.entitiesProcessorsNamesHolder
                .getFor(entityClass, actionNm);
              if (entProcNm == null) {
                this.secureLogger.error(null, HandlerEntityRequest.class,
                  "Trying to work with forbidden entity/action/user: " + nmEnt
                    + "/" + actionNm + "/" + pRqDt.getUserName());
                throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
                  "Forbidden!");
              }
              @SuppressWarnings("unchecked")
              IEntityProcessor<IHasId<?>, ?> ep =
                (IEntityProcessor<IHasId<?>, ?>)
                  this.entitiesProcessorsFactory.lazyGet(pRqVs, entProcNm);
              if (isShowDebMsg && detLev > 100) {
                this.logger.debug(pRqVs, HandlerEntityRequest.class,
                  "It's used entProcNm/IEntityProcessor: " + entProcNm
                    + "/" + ep.getClass());
              }
              entity = ep.process(pRqVs, entity, pRqDt);
            } else { // else actions like "list" (page)
              String procNm = this.processorsNamesHolder
                .getFor(entityClass, actionNm);
              if (procNm == null) {
                this.secureLogger.error(null, HandlerEntityRequest.class,
                "Trying to work with forbidden entity/action/user: " + nmEnt
                  + "/" + actionNm + "/" + pRqDt.getUserName());
                throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
                  "Forbidden!");
              }
              IProcessor proc = this.processorsFactory
                .lazyGet(pRqVs, procNm);
              if (isShowDebMsg && detLev > 100) {
                this.logger.debug(pRqVs, HandlerEntityRequest.class,
                  "It's used procNm/IProcessor: " + procNm + "/"
                    + proc.getClass());
              }
              proc.process(pRqVs, pRqDt);
            }
          }
          this.srvDatabase.commitTransaction();
        } catch (Exception ex) {
          this.srvDatabase.rollBackTransaction();
          throw ex;
        } finally {
          this.srvDatabase.releaseResources();
        }
      }
    } else {
      handleNoChangeIsolation(pRqVs, pRqDt, entityClass, actionsArr,
        isShowDebMsg, detLev, nmEnt);
    }
  }

  /**
   * <p>Handle request without changing transaction isolation.</p>
   * @param pRqVs Request scoped variables
   * @param pRqDt Request Data
   * @param pEntityClass Entity Class
   * @param pActionsArr Actions Array
   * @param pIsShowDebMsg Is Show Debug Messages
   * @param pDetLev Debug Messages detail level
   * @param pNmEnt entity name
   * @throws Exception - an exception
   */
  public final void handleNoChangeIsolation(final Map<String, Object> pRqVs,
    final IRequestData pRqDt, final Class<?> pEntityClass,
      final String[] pActionsArr, final boolean pIsShowDebMsg,
        final int pDetLev, final String pNmEnt) throws Exception {
    Class<?> entityClass = pEntityClass;
    try {
      this.srvDatabase.setIsAutocommit(false);
      this.srvDatabase.setTransactionIsolation(this.writeReTi);
      this.srvDatabase.beginTransaction();
      IHasId<?> entity = null;
      if (pActionsArr[0].startsWith("entity")) {
        // actions like "save", "delete"
        @SuppressWarnings("unchecked")
        IFactorySimple<IHasId<?>> entFac = (IFactorySimple<IHasId<?>>)
          this.entitiesFactoriesFatory.lazyGet(pRqVs, entityClass);
        entity = entFac.create(pRqVs);
        this.fillEntityFromReq.fill(pRqVs, entity, pRqDt);
      }
      for (String actionNm : pActionsArr) {
        if (actionNm.startsWith("entity")) {
          if (entity == null) { // it's may be change entity to owner:
           entity = (IHasId<?>) pRqVs.get("nextEntity");
           if (entity == null) {
              throw new ExceptionWithCode(ExceptionWithCode.WRONG_PARAMETER,
                "wrong_request_entity_not_filled");
            }
           entityClass = entity.getClass();
          }
          String entProcNm = this.entitiesProcessorsNamesHolder
            .getFor(entityClass, actionNm);
          if (entProcNm == null) {
            this.secureLogger.error(null, HandlerEntityRequest.class,
              "Trying to work with forbidden entity/action/user: " + pNmEnt
                + "/" + actionNm + "/" + pRqDt.getUserName());
            throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
              "Forbidden!");
          }
          @SuppressWarnings("unchecked")
          IEntityProcessor<IHasId<?>, ?> ep = (IEntityProcessor<IHasId<?>, ?>)
            this.entitiesProcessorsFactory.lazyGet(pRqVs, entProcNm);
          if (pIsShowDebMsg && pDetLev > 100) {
            this.logger.debug(pRqVs, HandlerEntityRequest.class,
              "It's used entProcNm/IEntityProcessor: " + entProcNm + "/"
                + ep.getClass());
          }
          entity = ep.process(pRqVs, entity, pRqDt);
        } else { // else actions like "list" (page)
          String procNm = this.processorsNamesHolder
            .getFor(entityClass, actionNm);
          if (procNm == null) {
            this.secureLogger.error(null, HandlerEntityRequest.class,
              "Trying to work with forbidden entity/action/user: " + pNmEnt
                + "/" + actionNm + "/" + pRqDt.getUserName());
            throw new ExceptionWithCode(ExceptionWithCode.FORBIDDEN,
              "Forbidden!");
          }
          IProcessor proc = this.processorsFactory.lazyGet(pRqVs, procNm);
          if (pIsShowDebMsg && pDetLev > 100) {
            this.logger.debug(pRqVs, HandlerEntityRequest.class,
            "It's used procNm/IProcessor: " + procNm + "/" + proc.getClass());
          }
          proc.process(pRqVs, pRqDt);
        }
      }
      this.srvDatabase.commitTransaction();
    } catch (Exception ex) {
      this.srvDatabase.rollBackTransaction();
      throw ex;
    } finally {
      this.srvDatabase.releaseResources();
    }
  }

  //Simple getters and setters:
  /**
   * <p>Geter for logger.</p>
   * @return ILogger
   **/
  public final ILogger getLogger() {
    return this.logger;
  }

  /**
   * <p>Setter for logger.</p>
   * @param pLogger reference
   **/
  public final void setLogger(final ILogger pLogger) {
    this.logger = pLogger;
  }

  /**
   * <p>Getter for srvDatabase.</p>
   * @return ISrvDatabase<RS>
   **/
  public final ISrvDatabase<RS> getSrvDatabase() {
    return this.srvDatabase;
  }

  /**
   * <p>Setter for srvDatabase.</p>
   * @param pSrvDatabase reference
   **/
  public final void setSrvDatabase(final ISrvDatabase<RS> pSrvDatabase) {
    this.srvDatabase = pSrvDatabase;
  }

  /**
   * <p>Getter for fillEntityFromReq.</p>
   * @return IFillerObjectsFrom<IRequestData>
   **/
  public final IFillerObjectsFrom<IRequestData> getFillEntityFromReq() {
    return this.fillEntityFromReq;
  }

  /**
   * <p>Setter for fillEntityFromReq.</p>
   * @param pFillEntityFromReq reference
   **/
  public final void setFillEntityFromReq(
    final IFillerObjectsFrom<IRequestData> pFillEntityFromReq) {
    this.fillEntityFromReq = pFillEntityFromReq;
  }

  /**
   * <p>Getter for entitiesFactoriesFatory.</p>
   * @return IFactoryAppBeansByClass<IFactorySimple<?>>
   **/
  public final IFactoryAppBeansByClass<IFactorySimple<?>>
    getEntitiesFactoriesFatory() {
    return this.entitiesFactoriesFatory;
  }

  /**
   * <p>Setter for entitiesFactoriesFatory.</p>
   * @param pEntitiesFactoriesFatory reference
   **/
  public final void setEntitiesFactoriesFatory(
    final IFactoryAppBeansByClass<IFactorySimple<?>> pEntitiesFactoriesFatory) {
    this.entitiesFactoriesFatory = pEntitiesFactoriesFatory;
  }

  /**
   * <p>Getter for entitiesProcessorsFactory.</p>
   * @return IFactoryAppBeansByName<IEntityProcessor>
   **/
  public final IFactoryAppBeansByName<IEntityProcessor>
    getEntitiesProcessorsFactory() {
    return this.entitiesProcessorsFactory;
  }

  /**
   * <p>Setter for entitiesProcessorsFactory.</p>
   * @param pEntitiesProcessorsFactory reference
   **/
  public final void setEntitiesProcessorsFactory(
    final IFactoryAppBeansByName<IEntityProcessor>
      pEntitiesProcessorsFactory) {
    this.entitiesProcessorsFactory = pEntitiesProcessorsFactory;
  }

  /**
   * <p>Getter for entitiesProcessorsNamesHolder.</p>
   * @return IHolderForClassByName<String>
   **/
  public final IHolderForClassByName<String>
    getEntitiesProcessorsNamesHolder() {
    return this.entitiesProcessorsNamesHolder;
  }

  /**
   * <p>Setter for entitiesProcessorsNamesHolder.</p>
   * @param pEntitiesProcessorsNamesHolder reference
   **/
  public final void setEntitiesProcessorsNamesHolder(
    final IHolderForClassByName<String> pEntitiesProcessorsNamesHolder) {
    this.entitiesProcessorsNamesHolder = pEntitiesProcessorsNamesHolder;
  }

  /**
   * <p>Getter for processorsFactory.</p>
   * @return IFactoryAppBeansByName<IProcessor>
   **/
  public final IFactoryAppBeansByName<IProcessor> getProcessorsFactory() {
    return this.processorsFactory;
  }

  /**
   * <p>Setter for processorsFactory.</p>
   * @param pProcessorsFactory reference
   **/
  public final void setProcessorsFactory(
    final IFactoryAppBeansByName<IProcessor> pProcessorsFactory) {
    this.processorsFactory = pProcessorsFactory;
  }

  /**
   * <p>Getter for processorsNamesHolder.</p>
   * @return IHolderForClassByName<String>
   **/
  public final IHolderForClassByName<String> getProcessorsNamesHolder() {
    return this.processorsNamesHolder;
  }

  /**
   * <p>Setter for processorsNamesHolder.</p>
   * @param pProcessorsNamesHolder reference
   **/
  public final void setProcessorsNamesHolder(
    final IHolderForClassByName<String> pProcessorsNamesHolder) {
    this.processorsNamesHolder = pProcessorsNamesHolder;
  }

  /**
   * <p>Getter for entitiesMap.</p>
   * @return Map<String, Class<?>>
   **/
  public final Map<String, Class<?>> getEntitiesMap() {
    return this.entitiesMap;
  }

  /**
   * <p>Setter for entitiesMap.</p>
   * @param pEntitiesMap reference
   **/
  public final void setEntitiesMap(final Map<String, Class<?>> pEntitiesMap) {
    this.entitiesMap = pEntitiesMap;
  }

  /**
   * <p>Getter for writeTi.</p>
   * @return Integer
   **/
  public final Integer getWriteTi() {
    return this.writeTi;
  }

  /**
   * <p>Setter for writeTi.</p>
   * @param pWriteTi reference
   **/
  public final void setWriteTi(final Integer pWriteTi) {
    this.writeTi = pWriteTi;
  }

  /**
   * <p>Getter for secureLogger.</p>
   * @return ILogger
   **/
  public final ILogger getSecureLogger() {
    return this.secureLogger;
  }

  /**
   * <p>Setter for secureLogger.</p>
   * @param pSecureLogger reference
   **/
  public final void setSecureLogger(final ILogger pSecureLogger) {
    this.secureLogger = pSecureLogger;
  }

  /**
   * <p>Getter for readTi.</p>
   * @return Integer
   **/
  public final Integer getReadTi() {
    return this.readTi;
  }

  /**
   * <p>Setter for readTi.</p>
   * @param pReadTi reference
   **/
  public final void setReadTi(final Integer pReadTi) {
    this.readTi = pReadTi;
  }

  /**
   * <p>Getter for writeReTi.</p>
   * @return Integer
   **/
  public final Integer getWriteReTi() {
    return this.writeReTi;
  }

  /**
   * <p>Setter for writeReTi.</p>
   * @param pWriteReTi reference
   **/
  public final void setWriteReTi(final Integer pWriteReTi) {
    this.writeReTi = pWriteReTi;
  }

  /**
   * <p>Getter for wrReSpTr.</p>
   * @return Boolean
   **/
  public final Boolean getWrReSpTr() {
    return this.wrReSpTr;
  }

  /**
   * <p>Setter for wrReSpTr.</p>
   * @param pWrReSpTr reference
   **/
  public final void setWrReSpTr(final Boolean pWrReSpTr) {
    this.wrReSpTr = pWrReSpTr;
  }

  /**
   * <p>Getter for delegateSt.</p>
   * @return IDelegate<IRequestData>
   **/
  public final IDelegate<IRequestData> getDelegateSt() {
    return this.delegateSt;
  }

  /**
   * <p>Setter for delegateSt.</p>
   * @param pDelegateSt reference
   **/
  public final void setDelegateSt(
    final IDelegate<IRequestData> pDelegateSt) {
    this.delegateSt = pDelegateSt;
  }
}
